
/*
 *  od.c
 *  Copyright (C) 1999 Brad Proctor
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

/*
 * $Id: od.c,v 1.3 1999/06/21 15:04:55 brad Exp $
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <ctype.h>
#include "version.h"

static char help_text[] =
{
   "Usage: od [-v][-A addr_base][-j skip][-N count][-t type][file...]\n"
   "       od [-bcdosx][file] [[+]offset[.][b]]\n\n"
   "  -A addr_base  Radix that the address is printed in\n"
   "     d, o, x    d = decimal, o = octal, x = hex. o is default\n"
   "  -c            Print bytes as characters\n"
   "  -d            Print words (2 bytes) in unsigned decimal\n"
   "  -j skip       Skip the first `skip' bytes\n"
   "  --help        Print this help message\n"
   "  --version     Print the program version\n\n"
   "Report bugs to fredlie@sprynet.com" 
};

static char str_table[][5] =
{
   " nul", " soh", " stx", " etx", " eot", " enq", " ack", " bel",
   "  bs", "  ht", "  nl", "  vt", "  ff", "  cr", "  so", "  si",
   " dle", " dc1", " dc2", " dc3", " dc4", " nak", " syn", " etb",
   " can", "  em", " sub", " esc", "  fs", "  gs", "  rs", "  us",
   "  sp",
};

static FILE *in = stdin;

static char flag_named = 0;
static char flag_chars = 0;
static char flag_dec = 0;
static int  count = 0;
static char addr_base[] = "%07o";
static char type[] = "%06o";

static void print_default(void)
{
   int b, c, i = 0;

   while(1)
   {
      c = getc(in);
      if(c == EOF)
         break;
      b = getc(in);
      if(b == EOF)
      {
         putchar(' ');
         printf(type, c);
         break;
      }
      c = (b << 8) + c;
      putchar(' ');
      printf(type, c);
      if(++i == 8)
      {
         putchar('\n');
         printf(addr_base, (int)ftell(in));
         i = 0;
      }
   }
}

static void print_named(void)
{
   int c, i = 0;
   while(1)
   {
      c = getc(in);
      if(c == EOF)
         break;
      if(c <= 040)
         fputs(str_table[c], stdout);
      else if(c == 0177)
         fputs(" del", stdout);
      else
         printf("   %c", c);
      if(++i == 16)
      {
         putchar('\n');
         printf(addr_base, (int)ftell(in));
         i = 0;
      }
   }
}

static void print_chars(void)
{
   int c, i = 0;

   while(1)
   {
      c = getc(in);
      if(c == EOF)
         break;
      switch(c)
      {
         case '\0': fputs("  \\0", stdout); break;
         case '\a': fputs("  \\a", stdout); break;
         case '\b': fputs("  \\b", stdout); break;
         case '\f': fputs("  \\f", stdout); break;
         case '\n': fputs("  \\n", stdout); break;
         case '\r': fputs("  \\r", stdout); break;
         case '\t': fputs("  \\t", stdout); break;
         case '\v': fputs("  \\v", stdout); break;
         default:
            if(isprint(c))
            {
               fputs("   ", stdout);
               fputc(c, stdout);
            }
            else
               printf(" %03o", c);
      }
      if(++i == 16)
      {
         putchar('\n');
         printf(addr_base, (int)ftell(in));
         i = 0;
      }
   }
}

static void print_u2(void)
{
   int b, c, i = 0;

   while(1)
   {
      c = getc(in);
      if(c == EOF)
         break;
      b = getc(in);
      if(b == EOF)
      {
         printf(" %5d", (unsigned short)c);
         break;
      } 
      c = (b << 8) + c;
      printf(" %5d", (unsigned short)c);
      if(++i == 8)
      {
         putchar('\n');
         printf(addr_base, (int)ftell(in));
         i = 0;
      }
   }
}

/*
 * Main
 */

int main(int argc, char **argv)
{
   int i, j = 0;
   char *ptr;
   
   for(i = 1; i < argc; ++i)
   {
      ptr = argv[i];
      if(*ptr == '-')
      {
         while(*++ptr)
         switch(*ptr)
         {
            case '-':
               ++ptr;
               if(!strcmp(ptr, "help"))
               {
                  puts(help_text);
                  return (0);
               }
               else if(!strcmp(ptr, "version"))
               {
                  puts("od (butils) "VERSION_STRING);
                  return (0);
               }
            case 'c':
               flag_chars = 1;
               break;
            case 'd':
               flag_dec = 1;
               break;
            case 'j':
               if(*++ptr != '\0')
               {
                  fprintf(stderr, "%s: Invalid option: `%s'", argv[0], 
                     argv[i]);
                  fputs("Try `od --help' for more information.\n", stderr);
                  return (1);
               }
               if(++i < argc)
               {
                  sscanf(argv[i], "%i", &j);
                  ptr = " ";
               }
               break;
            case 't':
               if(++i < argc)
               {
                  switch(argv[i][0])
                  {
                     case 'a':   flag_named = 1; break;
                     default:
                        fprintf(stderr, "%s: Invalid typestring.\n", argv[0]);
                        fputs("Try `od --help' for more information.\n", 
                           stderr);
                        return (1);
                  }
               }
               else
               {
                  fprintf(stderr, "%s: Missing typestring.\n", argv[0]);
                  fputs("Try `od --help' for more information.\n", stderr);
                  exit(1);
               }
               break;                     
            case 'A':
               if(++i < argc)
               {
                  if(argv[i][1] != '\0')
                  {
                     fprintf(stderr, "%s: Invalid address base `%s'", 
                        argv[0], argv[i]);
                     fputs("Try `od --help' for more information.\n", stderr);
                     return (1);
                  }
                  switch(argv[i][0])
                  {
                     case 'd': addr_base[3] = 'd'; break;
                     case 'o': addr_base[3] = 'o'; break;
                     case 'x': addr_base[3] = 'x'; break;
                     default:
                        fprintf(stderr, "%s: Invalid address base `%s'", 
                           argv[0], argv[i]);
                        fputs("Try `od --help' for more information.\n", 
                           stderr);
                        return (1);
                  }
               }
               break;
            case 'N':
               if(++i < argc)
                  sscanf(argv[i], "%i", &count);
               else
               {
                  fprintf(stderr, "%s: Argument expected after `-N'\n", 
                     argv[0]);
                  fputs("Try `od --help' for more information.\n", stderr);
                  return (1);
               }
               break;
            default:
               fprintf(stderr, "%s: Invalid option `%s'\n", argv[0], argv[i]);
               fputs("Try `od --help' for more information.", stderr);
               return (1);
         }
      }
      else
      {
         if((in = fopen(argv[i], "r")) == NULL)
         {
            fprintf(stderr, "%s: Could not open `%s' for reading.\n", 
               argv[0], argv[i]);
            fputs("Try `od --help' for more information.\n", stderr);
            return (1);
         }
         if(j)
         {
            fseek(in, j, SEEK_SET);
            j = 0;
         }
         printf(addr_base, 0);
         if(flag_named)
            print_named();
         else if(flag_chars)
            print_chars();
         else if(flag_dec)
            print_u2();
         else
            print_default();
         putchar('\n');
         printf(addr_base, (int)ftell(in));
         putchar('\n');
      }
   }
   return (0);
}